/** *****************  gestionCaptureImage.c ************************* **/
/*  Fouquet Jeremy	   Lefebvre Yann	Chevalier Henri		*/
/*	IMA 3I				2004/2005			*/
/************************************************************************/
/*	Adrian Siminiceanu			Liviu Amarfei		*/
/*	USM Suceava Roumanie		2005/2006			*/
/************************************************************************/
/*	Chamfroy Florent			Gourisse Maxime		*/
/*	IMA 3I				2005/2006			*/
/************************************************************************/

#include <errno.h>
#include <fcntl.h>

#include <linux/videodev.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <jpeglib.h>

#include "gestionServeurImage.h"
#include "gestionCaptureImage.h"

#include "../Threads/libthrd.h"


int descrCamera1;
int descrCamera2;
int read_img=0;

struct video_capability vid_caps;
struct video_window vid_win;
struct video_mbuf vid_buf;
struct jpeg_destination_mgr dstmgr;

static unsigned char tmp_buf[JPEG_BUF_SIZE];
static unsigned long int tmp_bytes;


/*Fonctions ncessaires pour l'initialisation du buffer de jpeg*/

static void jpeg_init_destination (struct jpeg_compress_struct *cinfo)
{
 cinfo->dest->next_output_byte=tmp_buf;
 cinfo->dest->free_in_buffer=JPEG_BUF_SIZE;
}

static int jpeg_empty_output_buffer(struct jpeg_compress_struct *cinfo)
{
 return 1;
}

static void jpeg_term_destination(struct jpeg_compress_struct *cinfo){}

void jpeg_compress(char *maptemp)
{
	int line_width,x,y;
	JSAMPROW row_ptr[1];
	struct jpeg_compress_struct jcomp;
	struct jpeg_error_mgr jerr;
	char line[WIDTH * 3];
	
	if (!line)return;
	jcomp.err = jpeg_std_error(&jerr);
	jpeg_create_compress (&jcomp);
	jcomp.image_width = WIDTH;
	jcomp.image_height= HEIGHT;
	jcomp.input_components = 3;
	jcomp.in_color_space = JCS_RGB;
	
	dstmgr.init_destination=jpeg_init_destination;
	dstmgr.empty_output_buffer=jpeg_empty_output_buffer;
	dstmgr.term_destination=jpeg_term_destination;
	jcomp.dest=&dstmgr;
	
	jpeg_set_defaults (&jcomp);
	jpeg_set_quality (&jcomp, QUALITY, TRUE);
	jcomp.dct_method = JDCT_FASTEST;
	
	jpeg_start_compress (&jcomp, TRUE);
		
	row_ptr[0] = line;
	
	line_width = WIDTH * 3;
	for ( y = 0; y < HEIGHT; y++) {
		for (x = 0; x < line_width; x+=3) {
			line[x]   = maptemp[x+2];
			line[x+1] = maptemp[x+1];
			line[x+2] = maptemp[x];
		}
		jpeg_write_scanlines (&jcomp, row_ptr, 1);
		maptemp += line_width;
	}
	
	jpeg_finish_compress (&jcomp);
	tmp_bytes=jcomp.dest->next_output_byte-tmp_buf;
	jpeg_destroy_compress (&jcomp);

	P(&mutex_jpeg_buf);
	memcpy(jpeg_buf,tmp_buf,JPEG_BUF_SIZE);
	jpeg_bytes=tmp_bytes;
	V(&mutex_jpeg_buf);
}

char *start_capture(int num_cam){
	char *maptemp;
	if (num_cam==1) {
		if (ioctl (descrCamera1, VIDIOCGCAP, &vid_caps) == -1) {
			perror ("ioctl (VIDIOCGCAP)");
			return (NULL);
		}
		if (ioctl (descrCamera1, VIDIOCGMBUF, &vid_buf) == -1) {
			fprintf(stderr, "no mmap falling back on read\n");
			if (ioctl (descrCamera1, VIDIOCGWIN, &vid_win)== -1) {
				perror ("ioctl VIDIOCGWIN");
				return (NULL);
			}
			vid_win.width=WIDTH;
			vid_win.height=HEIGHT;
			if (ioctl (descrCamera1, VIDIOCSWIN, &vid_win)== -1) {
				perror ("ioctl VIDIOCSWIN");
				return (NULL);
			}
			read_img=1;
			maptemp=(char *)malloc(WIDTH*HEIGHT*3);
			return (maptemp);
		}
		maptemp=mmap(0, vid_buf.size*3, PROT_READ|PROT_WRITE, MAP_SHARED, descrCamera1, 0);
	} else if (num_cam==2) {
		if (ioctl (descrCamera2, VIDIOCGCAP, &vid_caps) == -1) {
			perror ("ioctl (VIDIOCGCAP)");
			return (NULL);
		}
		if (ioctl (descrCamera2, VIDIOCGMBUF, &vid_buf) == -1) {
			fprintf(stderr, "no mmap falling back on read\n");
			if (ioctl (descrCamera2, VIDIOCGWIN, &vid_win)== -1) {
				perror ("ioctl VIDIOCGWIN");
				return (NULL);
			}
			vid_win.width=WIDTH;
			vid_win.height=HEIGHT;
			if (ioctl (descrCamera2, VIDIOCSWIN, &vid_win)== -1) {
				perror ("ioctl VIDIOCSWIN");
				return (NULL);
			}
			read_img=1;
			maptemp=(char *)malloc(WIDTH*HEIGHT*3);
			return (maptemp);
		}
		maptemp=mmap(0, vid_buf.size*3, PROT_READ|PROT_WRITE, MAP_SHARED, descrCamera2, 0);
	}

	if ((unsigned char *)-1 == (unsigned char *)maptemp)
		return (NULL);
	return (maptemp);
}


char *next_capture(char * maptemp, int num_cam){
	
	struct video_mmap vid_mmap1,vid_mmap2;

	if (read_img) {
		if (num_cam==1)
			if ((maptemp=mmap(0, vid_buf.size*3, PROT_READ|PROT_WRITE, MAP_SHARED, descrCamera1, 0)))
				return NULL;
		else if (num_cam==2)
			if ((maptemp=mmap(0, vid_buf.size*3, PROT_READ|PROT_WRITE, MAP_SHARED, descrCamera2, 0)))
				return NULL;
	} else {
		vid_mmap1.format=PALETTE;
		vid_mmap1.frame=0;
		vid_mmap1.width=WIDTH;
		vid_mmap1.height=HEIGHT;
		
		vid_mmap2.format=PALETTE;
		vid_mmap2.frame=0;
		vid_mmap2.width=WIDTH;
		vid_mmap2.height=HEIGHT;

		if (num_cam==1) {
			if (ioctl(descrCamera1, VIDIOCMCAPTURE, &vid_mmap1) == -1) {
				printf("%d Erreur VIDIOCMCAPTURE1\n",errno);
				return (NULL);
			}
			if (ioctl(descrCamera1, VIDIOCSYNC, &vid_mmap1) == -1) {
				printf("Erreur VIDIOCSYNC1\n");
				return (NULL);
			}
		} else if (num_cam==2) {
			if (ioctl(descrCamera2, VIDIOCMCAPTURE, &vid_mmap2) == -1) {
				printf("%d Erreur VIDIOCMCAPTURE2\n",errno);
				return (NULL);
			}
			if (ioctl(descrCamera2, VIDIOCSYNC, &vid_mmap2) == -1) {
				printf("Erreur VIDIOCSYNC2\n");
				return (NULL);
			}
		}
	}

	return maptemp;
}


void difference(char *map1, char *map2){
     int x,y;
     char *mapt1=map1, *mapt2=map2;
     int line_width=WIDTH*3;
	     for ( y = 0; y < HEIGHT; y++) {
		     for (x = 0; x < line_width; x+=3) {
			     mapt1[x]=mapt1[x]-mapt2[x];
			     mapt1[x+1]=mapt1[x+1]-mapt2[x+1];
			     mapt1[x+2]=mapt1[x+2]-mapt2[x+2];
		      }
		mapt1 += line_width;
		mapt2 += line_width;
	 }
}

int gestionCaptureImage()
{
	int bool=1;
	char *map1, *map2;
	descrCamera1=open("/dev/video1", O_RDWR);
	descrCamera2=open("/dev/video2", O_RDWR);

	
	if(descrCamera1 < 0){
		printf("\aAucun peripherique detecte\n");
		printf("Verifiez les droits d'acces a /dev/video1\n");
		exit(-1);
	}
	if(descrCamera2 < 0){
		printf("\aAucun peripherique detecte\n");
		printf("Verifiez les droits d'acces a /dev/video2\n");
		exit(-1);
	}
	
	map1=start_capture(1);
	map2=start_capture(2);
	if(!map1){
		printf("Failed to initialize webcam 1\n");
		exit(1);
	}
	if(!map2){
		printf("Failed to initialize webcam 2\n");
		exit(1);
	}

	map1=next_capture(map1, 1);
	map2=next_capture(map2, 2);
	if(!map1 || !map2) bool=0;
	while(bool){

difference(map1,map2);

		P(&mutex_nb_client);
		if(nb_client>0) jpeg_compress(map1);
		V(&mutex_nb_client);
		map1=next_capture(map1, 1);
		map2=next_capture(map2, 2);
		if(!map1 || !map2) bool=0;
		
	}
	
	close(descrCamera1);
	close(descrCamera2);
	munmap(map1, vid_buf.size*3);
	
	
	return (0);
}
