/** ***********************  gestionSon.c ****************************** **/
/*  Fouquet Jeremy	   Lefebvre Yann	Chevalier Henri		*/
/*	IMA 3I				2004/2005			*/
/************************************************************************/
/*	Adrian Siminiceanu			Liviu Amarfei		*/
/*	USM Suceava Roumanie		2005/2006			*/
/************************************************************************/
/*	Chamfroy Florent			Gourisse Maxime		*/
/*	IMA 3I				2005/2006			*/
/************************************************************************/

#include <getopt.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <linux/soundcard.h>

#include "gestionSon.h"
#include "centaure.h"

struct info_struct info;

static pthread_cond_t  cond;        /* Condition variable */
static pthread_mutex_t lock;        /* Mutex lock */

DWORD read_big_endian_long(byte *buffer)
{
#ifdef WORDS_BIGENDIAN
  DWORD byte0;
  memcpy(&byte0, buffer, 4);
  return byte0;
#else
  DWORD byte0, byte1, byte2, byte3;
  byte0 = (DWORD)buffer[0];
  byte1 = (DWORD)buffer[1];
  byte2 = (DWORD)buffer[2];
  byte3 = (DWORD)buffer[3];
  return(byte0 << 24 | byte1 << 16 | byte2 << 8 | byte3);
#endif
}

void write_big_endian_long(byte *buffer, DWORD value)
{
#ifdef WORDS_BIGENDIAN
  buffer[0] = (byte)(value     & 0xFF);
  buffer[1] = (byte)(value>>8  & 0xFF);
  buffer[2] = (byte)(value>>16 & 0xFF);
  buffer[3] = (byte)(value>>24 & 0xFF);
#else
  buffer[0] = (byte)(value>>24 & 0xFF);
  buffer[1] = (byte)(value>>16 & 0xFF);
  buffer[2] = (byte)(value>>8  & 0xFF);
  buffer[3] = (byte)(value     & 0xFF);
#endif
}

DWORD read_little_endian_long(byte *buffer)
{
#ifdef WORDS_BIGENDIAN
  DWORD byte0;
  memcpy(&byte0, buffer, 4);
  return ((byte0 >> 24) | ((byte0 >> 8) & 0xff00) |
               ((byte0 << 8) & 0xff0000) | (byte0 << 24));
#else
  DWORD byte0, byte1, byte2, byte3;
  byte0 = (DWORD)buffer[0];
  byte1 = (DWORD)buffer[1];
  byte2 = (DWORD)buffer[2];
  byte3 = (DWORD)buffer[3];
  return(byte3 << 24 | byte2 << 16 | byte1 << 8 | byte0);
#endif
}

WORD read_little_endian_word(byte *buffer)
{
#ifdef WORDS_BIGENDIAN
  WORD byte0;
  memcpy(&byte0, buffer, 2);
  return ((byte0 << 8) | (byte0 >> 8));
#else
  WORD byte0, byte1;
  byte0 = (WORD)buffer[0];
  byte1 = (WORD)buffer[1];
  return(byte1 << 8 | byte0);
#endif
}

static void init_info(){
  info.loop = FAUX;
  info.skip = FAUX;
  info.in_seconds = FAUX;
  info.speed    = DEFAULT_SPEED;
  info.bits     = DEFAULT_BITS;
  info.encoding = LIN_BITS;
  info.channels = DEFAULT_CHANNELS;
  info.buffer_size = AUDIO_BUFFER_SIZE;
  info.show_usage = FAUX;
  info.swap = FAUX;
  info.forceraw = FAUX;
  info.force = FAUX;
  info.device = NULL;
  info.verbose = FAUX;
  info.optind = 0;
  info.buffer = NULL;
  info.firstblock = NULL;
  info.number_of_blocks = 0;
  info.alldone = FAUX;
  info.audioset = FAUX;
  info.headerskip = 0;
  info.audio_fd = -1;
  info.file_fd = -1;
  info.blocksize = 0;
  info.bytes_on_last_block = 0;
  info.capacity = 0;
  info.writeblock = 0;
  info.data_length = 0;
  pthread_mutex_init(&lock, NULL);
  pthread_cond_init(&cond, NULL);
}

static void reset_info()
{
  info.firstblock = NULL;
  info.writeblock = 0;
  info.capacity = 0;
  info.alldone = FAUX;
  info.audioset = FAUX;
  info.headerskip = 0;
  info.bytes_on_last_block = 0;
  pthread_mutex_init(&lock, NULL);
  pthread_cond_init(&cond, NULL);
}


static void errdie(char *str){
  
  fprintf(stderr, "Error: %s\n", str);
  exit(-1);
}

static void allocate_buffer()
{
  if (info.buffer == NULL) {
    info.buffer = (byte*) malloc(info.buffer_size*sizeof(byte));
    if (info.buffer == NULL) 
      errdie("malloc failed");
    info.number_of_blocks = info.buffer_size / info.blocksize;
  }
}

static void calculate_buffer_size(int speed, int bits, int channels)
{
  int bytes_per_second;
  double seconds = 0.0;
  int tmp;

  bytes_per_second = speed * channels * (bits / 8);

  if (info.in_seconds) {
    tmp = (int)(1000.0 * info.seconds);
    info.buffer_size = bytes_per_second * tmp / 1000;
    seconds = info.seconds;
  } else {
    seconds = (double)info.buffer_size / (double)bytes_per_second;
  }

  if (info.buffer_size < 5*MIN_BLOCK_SIZE)
    info.buffer_size = 5*MIN_BLOCK_SIZE;

  if (info.verbose) 
    printf("Buffer size: %dk (%.1fs)\n", info.buffer_size / 1024, seconds);

}

static void calculate_skip_size(int speed, int bits, int channels)
{
  int bytes_per_second;

  if (info.skip) {
    bytes_per_second = speed * channels * (bits / 8);
    info.skip_size = bytes_per_second * info.sseconds;
    if (info.verbose) 
      printf("Skipping: %dk (%.1fs)\n", info.skip_size / 1024, info.sseconds);
    if (lseek(info.file_fd, (off_t)info.skip_size, SEEK_CUR) == (off_t)-1)
      printf("Failed to skip. Behavior may be erratic. Continuing anyway.\n");
  }
}

static int read_wav(byte *buffer)
{
  WORD format, channels, bits, encoding;
  DWORD magic, samples_per_second, length,data_length;

  magic              = read_little_endian_long(buffer);
  if (magic != RIFF_MAGIC) /* RIFF file? */
    return(1);
  magic              = read_little_endian_long(buffer+0x08);
  if (magic != WAVE_MAGIC) /* WAVE file? */
    return(1);
  magic              = read_little_endian_long(buffer+0x24);
  if ((magic != DATA_MAGIC) && (magic != INFO_MAGIC)) /* data-portion there? */
    {
      printf("Unknown WAV-header magic. Continuing anyway.");
    }
  length             = read_little_endian_long(buffer+0x10);
  /* Subchunk length should be 16 here */
  if (length != 16)
    errdie("Unknown RIFF/WAVE header");

  encoding = LIN_BITS;
  format             = read_little_endian_word(buffer+0x14);
  switch(format) {
  case 0x0001 : /* PCM format */
    // Already set
    break;
  case 0x0101 : /* mu-law */
    /* die("Mu-law RIFF/WAVE audio file not supported"); */
    encoding = ULAW_BITS;
    break;
  case 0x0102 : /* a-law */
    /* die("A-law RIFF/WAVE audio file not supported"); */
    encoding = ALAW_BITS;
    break;
  case 0x0103 : /* ADPCM */
    /* die("ADPCM RIFF/WAVE audio file not supported"); */
    encoding = ADPCM_BITS;
    break;
  default :
    errdie("Unknown RIFF/WAVE audio file format");
    break;
  }

  info.filetype = RIFF_FILE;

  channels           = read_little_endian_word(buffer+0x16);
  samples_per_second = read_little_endian_long(buffer+0x18);
  bits               = read_little_endian_word(buffer+0x22);

  if (bits == 12)
    errdie("12 bits per sample not supported");

  if (bits == 8 && encoding == LIN_BITS)
    encoding = LIN8_BITS;

  data_length        = read_little_endian_long(buffer+0x28);
  info.data_length   = read_little_endian_long(buffer+0x28);

  /* Set audio parameters */
  info.speed    = (int)samples_per_second;
  info.bits     = (int)bits;
  info.encoding = (int)encoding;
  info.channels = (int)channels;

  /* Move data to start from the beginning of the buffer.            */
  /* This is to ensure the correct behaviour of rounding when 16bits */
  /* and/or stereo sample is to be played.                           */
  memmove(buffer, buffer+0x2c, info.blocksize-0x2c);

  /* Save audio sample starting point */
  info.headerskip = 0x2c;

  return(0);
}

static void swap_block(byte *buffer, int blocksize)
{
  register int i;
  byte c, *p;
  p = buffer;
  for (i=0; i<blocksize/2; i++) {
    c = *p;
    *p = *(p+1);
    *++p = c;
    p++;
  }
}

static void read_header()
{
  int bytesread, count;
  byte *p, *bufferp;

  info.firstblock = (byte*) malloc(info.blocksize*sizeof(byte));
  if (info.firstblock == NULL) 
    errdie("malloc failed");

  /* Set bufferp to point to the beginning of buffer */
  bufferp = info.firstblock;
  /* If raw format is not forced, check header */
  if (!info.forceraw) {
    /* Read one block for header */
    bytesread = 0; 
    count = 0;
    p = bufferp;
    while ((bytesread < info.blocksize) && (count != -1) &&
	   ((count = read(info.file_fd, p, info.blocksize - bytesread)) != 0))
		{
		p+=count; 
		bytesread+=count;
		}

    read_wav(info.firstblock);
    
    /* Swap the block if needed. The header is swapped also but    */
    /* it is not needed anymore.                                   */
    if (info.swap)
      swap_block(bufferp, bytesread);

    /* If less bytes read than block size => all read => quit      */
    if (bytesread < info.blocksize) {
      info.alldone = VRAI;
      info.bytes_on_last_block = bytesread;
    }

    /* If header was found, read to the end of the block.          */
    if (info.headerskip) {
      count = 0;
      bytesread = info.blocksize - info.headerskip; 
      p = info.firstblock + (info.blocksize - info.headerskip);
      while ((bytesread < info.blocksize) && (count != -1) &&
	     ((count = read(info.file_fd, p, 
			    info.blocksize - bytesread)) != 0))
	{
	  p+=count; bytesread+=count;
	}
    }

    /* Advance to the next block */
    info.writeblock++;
    info.capacity++;

    /* We know the length of the data. Use it. */
    if (info.data_length) {
      if (info.data_length > info.blocksize)
	info.data_length -= info.blocksize;
      else
	{
	  /* We've already read all the data */
	  info.alldone = VRAI;
	  info.bytes_on_last_block = info.data_length;
	}
    }

  } 

  calculate_buffer_size(info.speed, info.bits, info.channels);
  allocate_buffer();
  calculate_skip_size(info.speed, info.bits, info.channels);
}


static void open_audio(){
  
  int block_size;
  char *device;
    
  device = AUDIO_DEVICE;
  
  if ((info.audio_fd = open(device, O_WRONLY, 0)) == -1)
    errdie("Cannot open audio device");

  ioctl(info.audio_fd, SNDCTL_DSP_GETBLKSIZE, &block_size);
  
  /* Audio device driver may give exotic audio buffer size     */
  /* for some adapters, we'd better set limits to it           */
  /* In the newer sound drivers, this should work according to */
  /* 4Front documents                                          */
  if (block_size < MIN_BLOCK_SIZE)
    block_size = MIN_BLOCK_SIZE;
  else if (block_size > MAX_BLOCK_SIZE)
    block_size = MAX_BLOCK_SIZE;

  /* Set audio block size and number of blocks */
  info.blocksize = block_size;
}

static void close_file(void)
{
  if ((info.file_fd != -1) && (info.file_fd != STDIN_FILENO)) {
    close(info.file_fd);
    info.file_fd = -1;
  }
}

static void open_file(char *file)
{
  close_file();
  if ((info.file_fd = open(file, O_RDONLY)) == -1) 
    errdie("Opening audio file failed");
}

static void destroy_buffer(void)
{
  /* Free memory */
  if (info.buffer != NULL)
    free(info.buffer);
}

static void reset_audio(void)
{
  if (ioctl(info.audio_fd, SNDCTL_DSP_RESET, 0) < 0)
    errdie("Audio reset failed");
}

static void sync_audio(void)
{
  if (ioctl(info.audio_fd, SNDCTL_DSP_SYNC, NULL) < 0)
    errdie("Audio sync failed");
}

static void close_audio(void)
{
  if (info.audio_fd != -1) {
    close(info.audio_fd);
    info.audio_fd = -1;
  }
}

static void set_audio_parameters()
{
  int format, stereo, stereo0, speed0;
  double percentage;

  reset_audio();

  stereo = 1;
  switch(info.channels) {
  case 1 :
    stereo = 0;
    break;
  case 2 :
    stereo = 1;
    break;
  default :
    if (!info.force)
      errdie("Invalid number of channels");
    break;
  }

  /* Set format */
  format = info.encoding;
  if (ioctl(info.audio_fd, SNDCTL_DSP_SETFMT, &format) == -1)
    errdie("SNDCTL_DSP_SETFMT");

  if (!info.force && (format != info.encoding))
    errdie("Audio device doesn't support this precision");

  /* Set stereo mode */
  stereo0 = stereo;
  if (ioctl(info.audio_fd, SNDCTL_DSP_STEREO, &stereo0) == -1)
    errdie("SNDCTL_DSP_STEREO");

  if (!info.force && (stereo0 != stereo))
    errdie("Audio device doesn't support stereo mode");

  info.channels = stereo0 + 1;

  /* Set sampling speed */
  speed0 = info.speed;
  if (ioctl(info.audio_fd, SNDCTL_DSP_SPEED, &speed0) == -1)
    errdie("SNDCTL_DSP_SPEED");

  /* Q: What to do if you cannot tune audio-card on exact speed? */
  /* Author: Jerko Golubovic <jerko.golubovic@public.srce.hr> */
  /* */
  /* Variant A: you force playing even if speed is not exact */
  /*   - Don't say anything - user choosed this way with force option */
  /*     We assume he/she knows what this means..... */
  /*   - Just skip and go ahead, jump :) (100.000 Lemmings can't be wrong :) )*/
  if (!info.force) { 
  /* Variant B: you don't force playing if speed is not exact */
     /* If speed doesn't match, die and inform user about alternative */
     if (speed0 != info.speed) { 
	fprintf(stderr, "Audio device doesn't support requested speed\n");
	fprintf(stderr, " -> Requested: %d Hz, Got: %d Hz", info.speed, speed0);
	percentage = ( (speed0-info.speed)/(double)info.speed ) * 100.0;
	fprintf(stderr, " (%+dHz/%+.2f%%) \n", speed0-info.speed, percentage);
	errdie("Use --force or -f in command-line to override.");
     }
     /*	If they match, you have:       */
     /*    a) good audio-card          */
     /*    b) reasonable requirements  */
     /* Just proceed and be happy about. Don't bother user with messages :) */
  }

  info.speed = speed0;
}

static void fill_buffer()
{
  int bytesread, count;
  byte *p, *bufferp;

  /* Move headerblock to buffer first */
  memmove(info.buffer, info.firstblock, info.blocksize);
  /* First block is not needed anymore so we free it */
  free(info.firstblock);

  bufferp = info.buffer + info.blocksize;
  while(VRAI) {

    /* Read a block piece by piece; especially if read from     */
    /* standard input, read may return unwanted block size.     */
    /* If count is -1, file_fd is invalid.                      */
    bytesread = 0; count = 0;
    p = bufferp;
    while ((bytesread < info.blocksize) && (count != -1) &&
	   ((count = read(info.file_fd, p, 
			  info.blocksize - bytesread)) != 0))
      {
	p+=count; bytesread+=count;
      }

    if (info.swap)
      swap_block(bufferp, bytesread);

    if (info.data_length) {
	if (info.data_length > info.blocksize)
	    info.data_length -= info.blocksize;
	else
	{
	    /* We've already read all the data */
	    info.last_block = info.writeblock;
	    info.bytes_on_last_block = info.data_length;
	    info.alldone = VRAI;
	    break;
	}
    }

    if (bytesread < info.blocksize) {
      info.last_block = info.writeblock;
      info.bytes_on_last_block = bytesread;
      info.alldone = VRAI;
      break;
    }

    /* Advance some pointers */
    bufferp+=info.blocksize;
    info.writeblock++;
    info.capacity++;
    if (info.writeblock >= info.number_of_blocks) {
      info.writeblock = 0;
      break;
    }
  }
}


static void * consumer(void *data)
{
  byte *bufferp;
  byte *bufferbase;
  int readblock, number_of_blocks;
  int audio_fd, blocksize, underflow;
  int lastblock, alldone, show_usage;

  /* Endless loop for the audio player */
  pthread_mutex_lock(&lock);
  audio_fd = info.audio_fd;
  blocksize = info.blocksize;
  number_of_blocks = info.number_of_blocks;
  bufferp = bufferbase = info.buffer;
  show_usage = info.show_usage;
  pthread_mutex_unlock(&lock);
  readblock = 0;
  for ( ; ; ) {
    
    pthread_mutex_lock(&lock);
    alldone = info.alldone;
    underflow = (info.capacity <= 0) ? VRAI : FAUX;
    lastblock = (info.last_block == readblock) ? VRAI : FAUX;
    audio_fd = info.audio_fd;
    pthread_mutex_unlock(&lock); 
   
    if (alldone) {
      /* If we hit the last block, write the last bytes and exit   */
      if (lastblock) {
	/* Write the last bytes and exit */
	/* info is not guarded by mutex here because producer      */
	/* should be exit by now.                                  */
	write(audio_fd, bufferp, info.bytes_on_last_block);
	break;
      }
    }

    if (underflow) {
      /* The audio buffer is empty; give producer a break */
      
      sync_audio();
      sleep(1);
      continue;
    }

    /* Push the audio data block to device */
    write(audio_fd, bufferp, blocksize);
    /*    printf("out: %d\n", readblock);*/

    /* Advance some pointers */
    bufferp+=blocksize;
    readblock++;
    if (readblock >= number_of_blocks) {
      bufferp = bufferbase;
      readblock = 0;
    }

    /* Decrease buffer capacity */
    pthread_mutex_lock(&lock);
    info.capacity--;
    pthread_mutex_unlock(&lock);

  }

  return NULL;
}

static void producer()
{
  int bytesread, count, writeblock;
  int overflow, file_fd, blocksize;
  byte *p, *bufferp, *bufferbase;
  int number_of_blocks,show_usage;

  /* Endless loop for the audio file reader and bufferer. */
  pthread_mutex_lock(&lock);
  bufferp = bufferbase = info.buffer;
  file_fd = info.file_fd;
  blocksize = info.blocksize;
  number_of_blocks = info.number_of_blocks;
  writeblock = info.writeblock;
  show_usage = info.show_usage;
  pthread_mutex_unlock(&lock);
  for ( ; ; ) {
    pthread_mutex_lock(&lock);
    overflow = (info.capacity >= info.number_of_blocks) ? VRAI : FAUX;
    pthread_mutex_unlock(&lock);

    if (overflow) {
      /* Overflow: Buffer is full */
      /* usleep(10000); Take a short nap or we will piss off CPU */
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 10000;
      select(0, NULL, NULL, NULL, &tv);
      continue; /* Loop back and try again */
    }

    /* Read a block piece by piece; especially if read from     */
    /* standard input, read may return unwanted block size.     */
    /* If count is -1, file_fd is invalid.                      */
    bytesread = 0; count = 0;
    p = bufferp;
    while ((bytesread < blocksize) && (count != -1) &&
	   ((count = read(file_fd, p, 
			  blocksize - bytesread)) != 0))
      {
	p+=count; bytesread+=count;
      }
    /*    printf("in: %d\n", writeblock);*/

    if (info.swap)
      swap_block(bufferp, bytesread);

    /* Use the length of the data if set. */
    if (info.data_length) {
      if (info.data_length > info.blocksize)
	info.data_length -= info.blocksize;
      else
	{
	  /* We've already read all the data */
	  pthread_mutex_lock(&lock);
	  info.last_block = writeblock;
	  info.bytes_on_last_block = info.data_length;
	  info.alldone = VRAI;
	  pthread_mutex_unlock(&lock);
	  break;
	}
    }

    /* If less than 'blocksize' read, we assume that this is          */
    /* the last block. Save data for consumer and exit this function. */
    /* If bytesread is -1, file_fd was invalid.                       */
    if (bytesread < blocksize) {
      pthread_mutex_lock(&lock);
      info.last_block = writeblock;
      info.bytes_on_last_block = bytesread;
      info.alldone = VRAI;
      pthread_mutex_unlock(&lock);
      break;
    }

    /* Advance some pointers */
    bufferp+=blocksize;
    writeblock++;
    if (writeblock >= number_of_blocks) {
      bufferp = bufferbase;
      writeblock = 0;
    }

    /* Increase buffer capacity */
    pthread_mutex_lock(&lock);
    info.capacity++;
    pthread_mutex_unlock(&lock);

  }

}

static void play(char *file)
{
  pthread_attr_t attr;
  pthread_t thread;
  void *retval;
  int err;

  #ifdef DEBUG
     printf("DEBUG : gestionSon:play : %s\n", file);
  #endif

  /* Reset flags */
  reset_info();

  open_audio();

  open_file(file);

  read_header();

  /* Set audio parameters */
  set_audio_parameters();

  /* Fill the audio buffer */
  fill_buffer();

  /* Set thread attributes: FIFO. */
  pthread_attr_init(&attr);
  pthread_attr_setschedpolicy(&attr, PTHREAD_CREATE_DETACHED);

  if ( (err=pthread_create(&thread, &attr, consumer, NULL)) ) {
    /* From Edward A. Falk: Only superuser can set SCHED_FIFO */
    /* so try pthread_create without attributes if failed.    */
    if( err != EPERM || pthread_create(&thread, &attr, consumer, NULL) )
      errdie("Cannot create thread");
    printf("No permissions for priority scheduling\n");
  }

  /* Start producing. Don't start if sample was smaller than  */
  /* audio buffer size (alldone flag is set)                  */
  if (!info.alldone)
    producer();

  /* Wait until the thread has finished. */
  pthread_join(thread, &retval);

  /* Close input file */
  close_file();
  close_audio();
  
}

static int test_audio_device(){

  char *device;
  int audio_test;

  device = AUDIO_DEVICE;

  if ((audio_test = open(device, O_WRONLY, 0)) == -1)
	{
	return 1;
	}
  else
	{
	close(audio_test);
	return 0;
	}
}

static int test_audio_file(char *file){

  int file_test;

  if ((file_test = open(file, O_RDONLY)) == -1)
        {
        return 1;
        }
  else
        {
        close(file_test);
        return 0;
        }
}


void gestionSon(char *file) {

  int audio_test = 0;  
  int file_test = 0;

  audio_test = test_audio_device();
  
  if(audio_test == 0) {
	
	  file_test = test_audio_file(file);

	  if(file_test == 0) {

	          init_info();

		  play(file);
  
		  destroy_buffer();
	  } else 
		   printf("DEBUG : gestionSon: Fichier %s non trouv !\n", file);
  } 
  else {
	  printf("DEBUG : gestionSon: Dispositif audio buch !\n");
          printf("DEBUG : gestionSon: Le son %s ne serra pas jou", file);
	}

}
