// compiler avec gcc glview.c -o glview -l freenect -lglut -O3
// ajouter -DVIDEO pour obtenir une visualisation vidéo

// installer : apt-get install freeglut3-dev
// utilisation scp : scp glview.c root@172.26.79.223:/home/pifou/Projet_S8_centaure/


//193.48.34.239
/*
* This file is part of the OpenKinect Project. http://www.openkinect.org
*
* Copyright (c) 2010 individual OpenKinect contributors. See the CONTRIB file
* for details.
*
* This code is licensed to you under the terms of the Apache License, version
* 2.0, or, at your option, the terms of the GNU General Public License,
* version 2.0. See the APACHE20 and GPL2 files for the text of the licenses,
* or the following URLs:
* http://www.apache.org/licenses/LICENSE-2.0
* http://www.gnu.org/licenses/gpl-2.0.txt
*
* If you redistribute this file in source form, modified or unmodified, you
* may:
* 1) Leave this header intact and distribute it under the same terms,
* accompanying it with the APACHE20 and GPL20 files, or
* 2) Delete the Apache 2.0 clause and accompany it with the GPL2 file, or
* 3) Delete the GPL v2 clause and accompany it with the APACHE20 file
* In all cases you must keep the copyright notice intact and include a copy
* of the CONTRIB file.
*
* Binary distributions must follow the binary distribution requirements of
* either License.
*/


#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "libfreenect.h"

#include <pthread.h>



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>


#define SOCKET_ECRITURE "/tmp/serial.send"
#define SOCKET_LECTURE "/tmp/serial.recv"


#if defined(__APPLE__)
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#include <math.h>
#include "centaure.h"



#define ZONE_INSENSIBLE_GAUCHE 40
#define ZONE_INSENSIBLE_DROITE 600
#define ZONE_INSENSIBLE_HAUT 100
#define TOLERENCE_ZONE_MORTE 50
#define SEUIL_DISTANCE_TROP_PROCHE 600
#define SEUIL_NB_POINT_TROP_PROCHE 30

pthread_t freenect_thread;
volatile int die = 0;

int g_argc;
char **g_argv;

int window;

pthread_mutex_t gl_backbuf_mutex = PTHREAD_MUTEX_INITIALIZER;

// back: owned by libfreenect (implicit for depth)
// mid: owned by callbacks, "latest frame ready"
// front: owned by GL, "currently being drawn"
uint8_t *depth_mid, *depth_front;
uint8_t *rgb_back, *rgb_mid, *rgb_front;

GLuint gl_depth_tex;
GLuint gl_rgb_tex;

freenect_context *f_ctx;
freenect_device *f_dev;
int freenect_angle = 0;
int freenect_led;

freenect_video_format requested_format = FREENECT_VIDEO_RGB;
freenect_video_format current_format = FREENECT_VIDEO_RGB;

pthread_cond_t gl_frame_cond = PTHREAD_COND_INITIALIZER;
int got_rgb = 0;
int got_depth = 0;

bool enableRecuperationZoneOmbre = true;
int gouleau = 0;
int se, sl; // interface série

int connexionServeur(char *chemin){
	int ds;                         /* Descripteur de SOCKET */
	struct sockaddr_un adresse;     /* Adresse du serveur */
	int statut;                     /* Stocke le statut des commandes */

	/* Creation de la SOCKET du client */
	ds=socket(AF_UNIX,SOCK_STREAM,0);
	if(ds<0) {perror("connexionServeur.socket"); exit(-1);}

	/* Preparation de la structure adresse du serveur */
	adresse.sun_family=AF_UNIX;
	strcpy(adresse.sun_path,chemin);

	/* Connexion au serveur */
	statut = connect(ds, (struct sockaddr *)&adresse, sizeof adresse);
	if(statut<0) {perror("connexionServeur.bind"); exit(-1);}

	return ds;
}

void DrawGLScene(){
	pthread_mutex_lock(&gl_backbuf_mutex);

	// When using YUV_RGB mode, RGB frames only arrive at 15Hz, so we shouldn't force them to draw in lock-step.
	// However, this is CPU/GPU intensive when we are receiving frames in lockstep.
	if (current_format == FREENECT_VIDEO_YUV_RGB) {
		while (!got_depth && !got_rgb) {
			pthread_cond_wait(&gl_frame_cond, &gl_backbuf_mutex);
		}
	} 
	else {
		while ((!got_depth || !got_rgb) && requested_format != current_format) {
			pthread_cond_wait(&gl_frame_cond, &gl_backbuf_mutex);
		}
	}

	if (requested_format != current_format) {
		pthread_mutex_unlock(&gl_backbuf_mutex);
		return;
	}

	uint8_t *tmp;

	if (got_depth) {
		tmp = depth_front;
		depth_front = depth_mid;
		depth_mid = tmp;
		got_depth = 0;
	}
	if (got_rgb) {
		tmp = rgb_front;
		rgb_front = rgb_mid;
		rgb_mid = tmp;
		got_rgb = 0;
	}

	pthread_mutex_unlock(&gl_backbuf_mutex);

	glBindTexture(GL_TEXTURE_2D, gl_depth_tex);
	glTexImage2D(GL_TEXTURE_2D, 0, 3, 640, 480, 0, GL_RGB, GL_UNSIGNED_BYTE, depth_front);

	glBegin(GL_TRIANGLE_FAN);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glTexCoord2f(0, 0); glVertex3f(0,0,0);
		glTexCoord2f(1, 0); glVertex3f(640,0,0);
		glTexCoord2f(1, 1); glVertex3f(640,480,0);
		glTexCoord2f(0, 1); glVertex3f(0,480,0);
	glEnd();

	glBindTexture(GL_TEXTURE_2D, gl_rgb_tex);
	
	if (current_format == FREENECT_VIDEO_RGB || current_format == FREENECT_VIDEO_YUV_RGB)
		glTexImage2D(GL_TEXTURE_2D, 0, 3, 640, 480, 0, GL_RGB, GL_UNSIGNED_BYTE, rgb_front);
	else
		glTexImage2D(GL_TEXTURE_2D, 0, 1, 640, 480, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, rgb_front+640*4);

	glBegin(GL_TRIANGLE_FAN);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glTexCoord2f(0, 0); glVertex3f(640,0,0);
		glTexCoord2f(1, 0); glVertex3f(1280,0,0);
		glTexCoord2f(1, 1); glVertex3f(1280,480,0);
		glTexCoord2f(0, 1); glVertex3f(640,480,0);
	glEnd();

	glutSwapBuffers();
}

void keyPressed(unsigned char key, int x, int y){
	if (key == 27) {
		die = 1;
		pthread_join(freenect_thread, NULL);
		glutDestroyWindow(window);
		free(depth_mid);
		free(depth_front);
		free(rgb_back);
		free(rgb_mid);
		free(rgb_front);
		// Not pthread_exit because OSX leaves a thread lying around and doesn't exit
		exit(0);
	}
	if (key == 'w') {
		freenect_angle++;
		if (freenect_angle > 30) {
			freenect_angle = 30;
		}
		freenect_set_tilt_degs(f_dev,freenect_angle);
	}
	if (key == 's') {
		freenect_angle = 0;
		freenect_set_tilt_degs(f_dev,freenect_angle);
	}
	if (key == 'f') {
		if (requested_format == FREENECT_VIDEO_IR_8BIT) requested_format = FREENECT_VIDEO_RGB;
		else if (requested_format == FREENECT_VIDEO_RGB) requested_format = FREENECT_VIDEO_YUV_RGB;
		else requested_format = FREENECT_VIDEO_IR_8BIT;
	}
	if (key == 'x') {
		freenect_angle--;
		if (freenect_angle < -30) {
			freenect_angle = -30;
		}
		freenect_set_tilt_degs(f_dev,freenect_angle);
	}
	
	if (key == 'y') {
		enableRecuperationZoneOmbre ^= 1;
		printf("\nRécupération des zones ombrées : %d\n", enableRecuperationZoneOmbre);
	}
	
	if (key == '1') {
		freenect_set_led(f_dev,LED_GREEN);
	}
	if (key == '2') {
		freenect_set_led(f_dev,LED_RED);
	}
	if (key == '3') {
		freenect_set_led(f_dev,LED_YELLOW);
	}
	if (key == '4') {
		freenect_set_led(f_dev,LED_BLINK_GREEN);
	}
	if (key == '5') {
		// 5 is the same as 4
		freenect_set_led(f_dev,LED_BLINK_GREEN);
	}
	if (key == '6') {
		freenect_set_led(f_dev,LED_BLINK_RED_YELLOW);
	}
	if (key == '0') {
		freenect_set_led(f_dev,LED_OFF);
	}
}

void ReSizeGLScene(int Width, int Height){
	glViewport(0,0,Width,Height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho (0, 1280, 480, 0, -1.0f, 1.0f);
	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void InitGL(int Width, int Height){
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glDepthFunc(GL_LESS);
		glDepthMask(GL_FALSE);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);
		glDisable(GL_ALPHA_TEST);
		glEnable(GL_TEXTURE_2D);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glShadeModel(GL_FLAT);

	glGenTextures(1, &gl_depth_tex);
	glBindTexture(GL_TEXTURE_2D, gl_depth_tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glGenTextures(1, &gl_rgb_tex);
	glBindTexture(GL_TEXTURE_2D, gl_rgb_tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	ReSizeGLScene(Width, Height);
}

void *gl_threadfunc(void *arg){
	printf("GL thread\n");

	glutInit(&g_argc, g_argv);

	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_ALPHA | GLUT_DEPTH);
	glutInitWindowSize(1280, 480);
	glutInitWindowPosition(0, 0);

	window = glutCreateWindow("LibFreenect");

	glutDisplayFunc(&DrawGLScene);
	glutIdleFunc(&DrawGLScene);
	glutReshapeFunc(&ReSizeGLScene);
	glutKeyboardFunc(&keyPressed);

	InitGL(1280, 480);

	glutMainLoop();

	return NULL;
}


uint16_t t_gamma[2048];
int cptGouleauCalculMoyennage = 0;

void depth_cb(freenect_device *dev, void *v_depth, uint32_t timestamp){
	int indexPixel;
	int i, j, k;
	int debutRupture, debutRuptureJ;
	bool zoneOmbre = false;
	int valeurRemplissage;
		
	int cmdStop;
	
	uint16_t *depth = (uint16_t*)v_depth;

	pthread_mutex_lock(&gl_backbuf_mutex);
	
/*
	if(gouleau==50){
		printf("depth : %d\n", depth[100]);
		gouleau=0;
	}
	else gouleau++;
*/

//*
	/////////////// Récupération des ombres ///////////////
	for (i=0; i<480; i++) {
		for(j=0; j<640; j++){
			indexPixel = 640*i+j;
			//printf("pval :%d depth : %d  lb : %d\n", pval, depth[indexPixel], lb);

			if(enableRecuperationZoneOmbre){
				if(depth[indexPixel]>=2040){	// le pixel est dans une zone d'ombre
					

	//				if(enableRecuperationZoneOmbre) depth[indexPixel] = depth[indexPixel-1];
					//depth[indexPixel] = depth[indexPixel-1];
					
					if(!zoneOmbre){		// détection du début d'une ombre
						debutRupture = indexPixel;
						debutRuptureJ = j;
						zoneOmbre = true;
					}

					if(zoneOmbre && j==639){					
						for(k=debutRupture; k<indexPixel; k++){
							depth[k] = depth[debutRupture-2];
							zoneOmbre = false;
						}						
					}
				
				}
				else if(debutRuptureJ>6){	// le pixel n'est pas dans une zone d'ombre
	
					if(zoneOmbre && j!=639){	// détection d'une fin d'ombre
						valeurRemplissage = (depth[debutRupture-2] > depth[indexPixel]) ? depth[debutRupture-2] : depth[indexPixel];
					
						//printf("remplissage avec : %d de %d à %d\n", valeurRemplissage, debutRupture, indexPixel);
						for(k=debutRupture; k<indexPixel; k++){
							depth[k] = valeurRemplissage;
						}					
					}

					zoneOmbre = false;
				}
				else if(debutRuptureJ<=6 ){

					if(zoneOmbre && j!=639){
						//printf("debutRuptureJ : %d", debutRuptureJ);						
					
						for(k=debutRupture; k<indexPixel; k++){
							depth[k] = depth[indexPixel];
						}					
					}

					zoneOmbre = false;
				}
			}


		}
	}
//*/

			
	
	#ifdef VIDEO
	/////////////// Affichage ///////////////
//*
	for (i=0; i<480; i++) {
		for(j=0; j<640; j++){
			indexPixel = 640*i+j;
			int pval = t_gamma[depth[indexPixel]];
			int lb = pval & 0xff; 
			switch (pval>>8) {
			case 0:
			depth_mid[3*indexPixel+0] = 255;
			depth_mid[3*indexPixel+1] = 255-lb;
			depth_mid[3*indexPixel+2] = 255-lb;
			break;
			case 1:
			depth_mid[3*indexPixel+0] = 255;
			depth_mid[3*indexPixel+1] = lb;
			depth_mid[3*indexPixel+2] = 0;
			break;
			case 2:
			depth_mid[3*indexPixel+0] = 255-lb;
			depth_mid[3*indexPixel+1] = 255;
			depth_mid[3*indexPixel+2] = 0;
			break;
			case 3:
			depth_mid[3*indexPixel+0] = 0;
			depth_mid[3*indexPixel+1] = 255;
			depth_mid[3*indexPixel+2] = lb;
			break;
			case 4:
			depth_mid[3*indexPixel+0] = 0;
			depth_mid[3*indexPixel+1] = 255-lb;
			depth_mid[3*indexPixel+2] = 255;
			break;
			case 5:
			depth_mid[3*indexPixel+0] = 0;
			depth_mid[3*indexPixel+1] = 0;
			depth_mid[3*indexPixel+2] = 255-lb;
			break;
			default:
			depth_mid[3*indexPixel+0] = 0;
			depth_mid[3*indexPixel+1] = 0;
			depth_mid[3*indexPixel+2] = 0;
			break;

			}

			
		}
	}
	
	/////////////// Affichage zones insensibles ///////////////
	j = ZONE_INSENSIBLE_GAUCHE;
	for(i=0; i<480; i++){		
		indexPixel = 640*i+j;
		depth_mid[3*indexPixel+0] = 255;
		depth_mid[3*indexPixel+1] = 0;
		depth_mid[3*indexPixel+2] = 0;
	}
	j = ZONE_INSENSIBLE_DROITE;
	for(i=0; i<480; i++){		
		indexPixel = 640*i+j;
		depth_mid[3*indexPixel+0] = 255;
		depth_mid[3*indexPixel+1] = 0;
		depth_mid[3*indexPixel+2] = 0;
	}
	i = ZONE_INSENSIBLE_HAUT;
	for (j=0; j<640; j++) {
		indexPixel = 640*i+j;
		depth_mid[3*indexPixel+0] = 255;
		depth_mid[3*indexPixel+1] = 0;
		depth_mid[3*indexPixel+2] = 0;
	}
	#endif
//*/
	//*
	/////////////// Moyennages des zones et décision ///////////////
	cptGouleauCalculMoyennage ++;		
	int nbPointTropProcheD = 0;
	int nbPointTropProcheG = 0;
	int sumG=0;
	int sumD=0;
	int moyG, moyD, diff;
	float tensionG, tensionD;
	float frein;
	unsigned char messageSerie[5];

	if(cptGouleauCalculMoyennage == 20){
		cptGouleauCalculMoyennage = 0;
		
		// Moyennage
		for (i=ZONE_INSENSIBLE_HAUT; i<480; i++) {
			for(j=ZONE_INSENSIBLE_GAUCHE; j<320; j++){
				indexPixel = 640*i+j;
				sumG += depth[indexPixel];
				
				if( depth[indexPixel] < SEUIL_DISTANCE_TROP_PROCHE ) nbPointTropProcheG++;
			}
		}
		for (i=ZONE_INSENSIBLE_HAUT; i<480; i++) {
			for(j=320; j<ZONE_INSENSIBLE_DROITE; j++){
				indexPixel = 640*i+j;
				sumD += depth[indexPixel];
				
				if( depth[indexPixel] < SEUIL_DISTANCE_TROP_PROCHE ) nbPointTropProcheD++;
			}
		}
		
		moyG = sumG/( (480-ZONE_INSENSIBLE_HAUT)*(320-ZONE_INSENSIBLE_GAUCHE) );
		moyD = sumD/( (480-ZONE_INSENSIBLE_HAUT)*(ZONE_INSENSIBLE_DROITE-320) );
		diff = moyG - moyD;
		
		// Décision
		#ifdef DEBUG
		printf("\n\nbPointTropProcheG : %d  nbPointTropProcheD : %d\n", nbPointTropProcheG, nbPointTropProcheD);
		printf("Valeur diff : %d  moy gauche : %d moy droite : %d\n", diff, moyG, moyD);
		#endif
		if(nbPointTropProcheG > SEUIL_NB_POINT_TROP_PROCHE && nbPointTropProcheD > SEUIL_NB_POINT_TROP_PROCHE){
			#ifdef DEBUG
			printf("Stop\n");
			#endif
			tensionG = 0;
			tensionD = 0;
		}
		else if(nbPointTropProcheG > SEUIL_NB_POINT_TROP_PROCHE){
			#ifdef DEBUG
			printf("Droite urgence !\n");
			#endif
			tensionG = 6;
			tensionD = 0;
		}
		else if(nbPointTropProcheD > SEUIL_NB_POINT_TROP_PROCHE){
			#ifdef DEBUG
			printf("Gauche urgence !\n");
			#endif
			tensionG = 0;
			tensionD = 6;
		}
		else{
			if( diff > TOLERENCE_ZONE_MORTE ){
				#ifdef DEBUG
				printf("Gauche\n");
				#endif
				tensionG = 2;
				tensionD = 8;
			}
			else if( diff < -TOLERENCE_ZONE_MORTE ){
				#ifdef DEBUG
				printf("Droite\n");
				#endif
				tensionG = 8;
				tensionD = 2;
			}
			else{
				#ifdef DEBUG
				printf("Tout droit\n");
				#endif
				write(tubeIS[1],ARRET,ARRET_SIZE);
				tensionG = 10;
				tensionD = 10;
			}
		}
	}
//*/

	// écriture sur le port série
	messageSerie[0] = tensionG;
	messageSerie[1] = tensionD;
	messageSerie[2] = frein;
//	if( write(se, messageSerie, 5*sizeof(char)) != 1 ){ perror("ecriture serie"); exit(-1); }

	got_depth++;
	pthread_cond_signal(&gl_frame_cond);
	pthread_mutex_unlock(&gl_backbuf_mutex);
}

void rgb_cb(freenect_device *dev, void *rgb, uint32_t timestamp){
	pthread_mutex_lock(&gl_backbuf_mutex);

	// swap buffers
	assert (rgb_back == rgb);
	rgb_back = rgb_mid;
	freenect_set_video_buffer(dev, rgb_back);
	rgb_mid = (uint8_t*)rgb;

	got_rgb++;
	pthread_cond_signal(&gl_frame_cond);
	pthread_mutex_unlock(&gl_backbuf_mutex);
}

void *freenect_threadfunc(void *arg){
	int accelCount = 0;

	freenect_set_tilt_degs(f_dev,freenect_angle);
	freenect_set_led(f_dev,LED_RED);
	freenect_set_depth_callback(f_dev, depth_cb);
	#ifdef VIDEO
	freenect_set_video_callback(f_dev, rgb_cb);
	freenect_set_video_mode(f_dev, freenect_find_video_mode(FREENECT_RESOLUTION_LOW, current_format));
	freenect_set_video_buffer(f_dev, rgb_back);
	freenect_start_video(f_dev);
	#endif
	freenect_set_depth_mode(f_dev, freenect_find_depth_mode(FREENECT_RESOLUTION_LOW, FREENECT_DEPTH_11BIT));

	freenect_start_depth(f_dev);

	printf("'w'-tilt up, 's'-level, 'x'-tilt down, '0'-'6'-select LED mode, 'f'-video format\n");

	while (!die && freenect_process_events(f_ctx) >= 0) {
		//Throttle the text output
/*
			if (accelCount++ >= 2000){


			accelCount = 0;
			freenect_raw_tilt_state* state;
			freenect_update_tilt_state(f_dev);
			state = freenect_get_tilt_state(f_dev);
			double dx,dy,dz;
			freenect_get_mks_accel(state, &dx, &dy, &dz);
			printf("\r raw acceleration: %4d %4d %4d mks acceleration: %4f %4f %4f", state->accelerometer_x, state->accelerometer_y, state->accelerometer_z, dx, dy, dz);
			fflush(stdout);
		}
*/

		#ifdef VIDEO
		if (requested_format != current_format) {
			freenect_stop_video(f_dev);
			freenect_set_video_mode(f_dev, freenect_find_video_mode(FREENECT_RESOLUTION_MEDIUM, requested_format));
			freenect_start_video(f_dev);
			current_format = requested_format;
		}
		#endif
//*/
	}

	printf("\nshutting down streams...\n");

	freenect_stop_depth(f_dev);
	freenect_stop_video(f_dev);

	freenect_close_device(f_dev);
	freenect_shutdown(f_ctx);

	printf("-- done!\n");
	return NULL;
}

void fermeture(){
	close(sl);
	close(se);
}

int main(int argc, char **argv){
	int res;

	atexit(fermeture);

	depth_mid = (uint8_t*)malloc(640*480*3);
	depth_front = (uint8_t*)malloc(640*480*3);
	#ifdef VIDEO
	rgb_back = (uint8_t*)malloc(640*480*3);
	rgb_mid = (uint8_t*)malloc(640*480*3);
	rgb_front = (uint8_t*)malloc(640*480*3);
	#endif
/*
	int c=0;
	int se=connexionServeur(SOCKET_ECRITURE);
	if(se<0){ fprintf(stderr,"Serveur serie (ecriture) absent\\n"); exit(-1); }
	int sl=connexionServeur(SOCKET_LECTURE);
	if(sl<0){ fprintf(stderr,"Serveur serie (electure) absent\\n"); exit(-1); }
*/
/*
	int i;
	for(i=0;i<8;i++){
	  if(read(sl,&c,sizeof(char))!=1){ perror("main.read"); exit(-1); }
	  printf("%02x\n",c);
  	}
*/

	printf("Kinect camera test\n");
//*
	int i;
	for (i=0; i<2048; i++) {
	float v = i/2048.0;
	v = powf(v, 3)* 6;
	t_gamma[i] = v*6*256;
	}
//*/
	g_argc = argc;
	g_argv = argv;

	if (freenect_init(&f_ctx, NULL) < 0) {
	printf("freenect_init() failed\n");
	return 1;
	}

	freenect_set_log_level(f_ctx, FREENECT_LOG_DEBUG);
	freenect_select_subdevices(f_ctx, (freenect_device_flags)(FREENECT_DEVICE_MOTOR | FREENECT_DEVICE_CAMERA));

	int nr_devices = freenect_num_devices (f_ctx);
	printf ("Number of devices found: %d\n", nr_devices);

	int user_device_number = 0;
	if (argc > 1)
	user_device_number = atoi(argv[1]);

	if (nr_devices < 1) {
	freenect_shutdown(f_ctx);
	return 1;
	}

	if (freenect_open_device(f_ctx, &f_dev, user_device_number) < 0) {
	printf("Could not open device\n");
	freenect_shutdown(f_ctx);
	return 1;
	}

	res = pthread_create(&freenect_thread, NULL, freenect_threadfunc, NULL);
	if (res) {
	printf("pthread_create failed\n");
	freenect_shutdown(f_ctx);
	return 1;
	}

	// OS X requires GLUT to run on the main thread
	#ifdef VIDEO
	gl_threadfunc(NULL);
	#endif
	while(1);

	return 0;
}
